// Enchanter
// Copyright (C) 2018 Katayama Hirofumi MZ <katayama.hirofumi.mz@gmail.com>
// This file is public domain software.
#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <cstdlib>
#include <cstring>

//static const TCHAR s_szGhostClass[] = TEXT("Ghost");
static const TCHAR s_szGhostClass[] = TEXT("My Ghost");

HWND g_hwndEnchanted = NULL;
HWND g_hwndGhost = NULL;
TCHAR g_szEnchantedText[MAX_PATH] = { 0 };
HBITMAP g_hbmEnchanted = NULL;

HBITMAP Create32BppBitmap(INT cx, INT cy)
{
    BITMAPINFO bi;
    ZeroMemory(&bi, sizeof(bi));
    bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bi.bmiHeader.biWidth = cx;
    bi.bmiHeader.biHeight = cy;
    bi.bmiHeader.biPlanes = 1;
    bi.bmiHeader.biBitCount = 32;
    HBITMAP hbm = NULL;
    if (HDC hdc = CreateCompatibleDC(NULL))
    {
        LPVOID pvBits;
        hbm = CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, &pvBits, NULL, 0);
        DeleteDC(hdc);
    }
    return hbm;
}

void DoUnenchant(BOOL bDestroyOriginal = FALSE)
{
    DeleteBitmap(g_hbmEnchanted);
    g_hbmEnchanted = NULL;
    if (bDestroyOriginal)
    {
        DWORD pid;
        GetWindowThreadProcessId(g_hwndEnchanted, &pid);
        if (HANDLE hProcess = OpenProcess(PROCESS_TERMINATE, FALSE, pid))
        {
            TerminateProcess(hProcess, -1);
            CloseHandle(hProcess);
        }

        DestroyWindow(g_hwndEnchanted);
        g_hwndEnchanted = NULL;
    }
    else
    {
        DestroyWindow(g_hwndGhost);
        g_hwndGhost = NULL;
        ShowWindowAsync(g_hwndEnchanted, SW_SHOWNOACTIVATE);
    }
}

typedef struct tagBITMAPINFOEX
{
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD          bmiColors[256];
} BITMAPINFOEX, FAR * LPBITMAPINFOEX;

BOOL SaveBitmapToFile(LPCTSTR pszFileName, HBITMAP hbm)
{
    BOOL f;
    BITMAPFILEHEADER bf;
    BITMAPINFOEX bi;
    BITMAPINFOHEADER *pbmih;
    DWORD cb;
    DWORD cColors, cbColors;
    HDC hDC;
    HANDLE hFile;
    LPVOID pBits;
    BITMAP bm;
    DWORD dwError = 0;

    if (!GetObject(hbm, sizeof(BITMAP), &bm))
        return FALSE;

    pbmih = &bi.bmiHeader;
    ZeroMemory(pbmih, sizeof(BITMAPINFOHEADER));
    pbmih->biSize             = sizeof(BITMAPINFOHEADER);
    pbmih->biWidth            = bm.bmWidth;
    pbmih->biHeight           = bm.bmHeight;
    pbmih->biPlanes           = 1;
    pbmih->biBitCount         = bm.bmBitsPixel;
    pbmih->biCompression      = BI_RGB;
    pbmih->biSizeImage        = bm.bmWidthBytes * bm.bmHeight;

    if (bm.bmBitsPixel < 16)
        cColors = 1 << bm.bmBitsPixel;
    else
        cColors = 0;
    cbColors = cColors * sizeof(RGBQUAD);

    bf.bfType = 0x4d42;
    bf.bfReserved1 = 0;
    bf.bfReserved2 = 0;
    cb = sizeof(BITMAPFILEHEADER) + pbmih->biSize + cbColors;
    bf.bfOffBits = cb;
    bf.bfSize = cb + pbmih->biSizeImage;

    pBits = HeapAlloc(GetProcessHeap(), 0, pbmih->biSizeImage);
    if (pBits == NULL)
        return FALSE;

    f = FALSE;
    hDC = GetDC(NULL);
    if (hDC != NULL)
    {
        if (GetDIBits(hDC, hbm, 0, bm.bmHeight, pBits, (BITMAPINFO*)&bi,
            DIB_RGB_COLORS))
        {
            hFile = CreateFile(pszFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                               CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL |
                               FILE_FLAG_WRITE_THROUGH, NULL);
            if (hFile != INVALID_HANDLE_VALUE)
            {
                f = WriteFile(hFile, &bf, sizeof(BITMAPFILEHEADER), &cb, NULL) &&
                    WriteFile(hFile, &bi, sizeof(BITMAPINFOHEADER), &cb, NULL) &&
                    WriteFile(hFile, bi.bmiColors, cbColors, &cb, NULL) &&
                    WriteFile(hFile, pBits, pbmih->biSizeImage, &cb, NULL);
                if (!f)
                    dwError = GetLastError();
                CloseHandle(hFile);

                if (!f)
                    DeleteFile(pszFileName);
            }
            else
                dwError = GetLastError();
        }
        else
            dwError = GetLastError();
        ReleaseDC(NULL, hDC);
    }
    else
        dwError = GetLastError();

    HeapFree(GetProcessHeap(), 0, pBits);
    SetLastError(dwError);
    return f;
}

VOID Whiten32BppBitmap(HBITMAP hbm)
{
    BITMAP bm;
    GetObject(hbm, sizeof(bm), &bm);

    if (bm.bmBitsPixel != 32 || !bm.bmBits)
        return;

    DWORD *pdw = (DWORD *)bm.bmBits;
    for (DWORD i = 0; i < bm.bmWidth * bm.bmHeight; ++i)
    {
        *pdw = *pdw | 0x00C0C0C0;
        ++pdw;
    }
}

void OnDraw(HDC hdc)
{
    BITMAP bm;
    GetObject(g_hbmEnchanted, sizeof(bm), &bm);

    if (HDC hdcMem = CreateCompatibleDC(hdc))
    {
        HGDIOBJ hbmOld = SelectObject(hdcMem, g_hbmEnchanted);
        {
            BitBlt(hdc, 0, 0, bm.bmWidth, bm.bmHeight, hdcMem, 0, 0, SRCCOPY | CAPTUREBLT);
        }
        SelectObject(hdcMem, hbmOld);
        DeleteDC(hdcMem);
    }
}

void OnNCPaint(HWND hwnd, HRGN hrgn)
{
    DefWindowProc(hwnd, WM_NCPAINT, (WPARAM)hrgn, 0);
    if (HDC hdc = GetWindowDC(hwnd))
    {
        OnDraw(hdc);
        ReleaseDC(hwnd, hdc);
    }
}

BOOL OnEraseBkgnd(HWND hwnd, HDC hdc)
{
    OnNCPaint(hwnd, NULL);
    return TRUE;
}

void OnPaint(HWND hwnd)
{
    PAINTSTRUCT ps;
    if (HDC hdc = BeginPaint(hwnd, &ps))
    {
        EndPaint(hwnd, &ps);
    }
}

void OnMove(HWND hwnd, int x, int y)
{
    RECT rc;
    GetWindowRect(hwnd, &rc);

    SetWindowPos(g_hwndEnchanted, NULL, rc.left, rc.top, 0, 0,
        SWP_NOSIZE | SWP_NOOWNERZORDER | SWP_NOSENDCHANGING | SWP_NOACTIVATE);
}

void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
    RECT rc;
    GetWindowRect(hwnd, &rc);

    SetWindowPos(g_hwndEnchanted, NULL, 0, 0, rc.right - rc.left, rc.bottom - rc.top,
        SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOSENDCHANGING | SWP_NOACTIVATE);
}

void OnDestroy(HWND hwnd)
{
    KillTimer(hwnd, 999);
}

void OnTimer(HWND hwnd, UINT id)
{
    if (id == 999)
    {
        KillTimer(hwnd, id);
        if (!IsWindow(g_hwndEnchanted))
        {
            // resume if window is destroyed
            DoUnenchant();
            return;
        }
        if (0)
        {
            // resume if responding
            DWORD_PTR dwResult;
            DWORD dwTimeout = 200;
            DWORD dwTick1 = GetTickCount();
            SendMessageTimeout(g_hwndEnchanted, WM_NULL, 0, 0, 0, dwTimeout, &dwResult);
            DWORD dwTick2 = GetTickCount();
            if (dwTick2 - dwTick1 < dwTimeout)
            {
                DoUnenchant();
                return;
            }
        }
        SetTimer(g_hwndGhost, 999, 1000, NULL);
    }
}

void OnClose(HWND hwnd)
{
    KillTimer(hwnd, 999);
    if (IDYES == MessageBox(hwnd, TEXT("Are you close this window?"),
                            TEXT("Hunging"), MB_ICONINFORMATION | MB_YESNO))
    {
        DoUnenchant(TRUE);
        return;
    }
    SetTimer(g_hwndGhost, 999, 1000, NULL);
}

void OnNCMouseMove(HWND hwnd, int x, int y, UINT codeHitTest)
{
    FORWARD_WM_NCMOUSEMOVE(hwnd, x, y, codeHitTest, DefWindowProc);
    SetCursor(LoadCursor(NULL, IDC_WAIT));
}

LRESULT CALLBACK
GhostWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        HANDLE_MSG(hwnd, WM_NCPAINT, OnNCPaint);
        HANDLE_MSG(hwnd, WM_ERASEBKGND, OnEraseBkgnd);
        HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
        HANDLE_MSG(hwnd, WM_MOVE, OnMove);
        HANDLE_MSG(hwnd, WM_SIZE, OnSize);
        HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
        HANDLE_MSG(hwnd, WM_CLOSE, OnClose);
        HANDLE_MSG(hwnd, WM_TIMER, OnTimer);
        HANDLE_MSG(hwnd, WM_NCMOUSEMOVE, OnNCMouseMove);
    }
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

HBITMAP GetWindowBitmap(HWND hwnd, INT cx, INT cy)
{
    HBITMAP hbm = NULL;
    if (HDC hdc = GetWindowDC(hwnd))
    {
        hbm = Create32BppBitmap(cx, cy);
        if (hbm)
        {
            if (HDC hdcMem = CreateCompatibleDC(hdc))
            {
                HGDIOBJ hbmOld = SelectObject(hdcMem, hbm);
                BitBlt(hdcMem, 0, 0, cx, cy, hdc, 0, 0, SRCCOPY | CAPTUREBLT);
                SelectObject(hdcMem, hbmOld);
                DeleteDC(hdcMem);
            }
        }
        ReleaseDC(hwnd, hdc);
    }
    return hbm;
}

BOOL DoEnchant(HWND hwndTarget)
{
    WNDCLASS wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.style = 0;
    wc.lpfnWndProc = GhostWindowProc;
    wc.hInstance = GetModuleHandle(NULL);
    wc.hIcon = NULL;
    wc.hCursor = LoadCursor(NULL, IDC_WAIT);
    wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
    wc.lpszClassName = s_szGhostClass;
    RegisterClass(&wc);

    RECT rc;
    GetWindowRect(hwndTarget, &rc);
    GetWindowText(hwndTarget, g_szEnchantedText, ARRAYSIZE(g_szEnchantedText));
    DWORD style = GetWindowStyle(hwndTarget);
    DWORD exstyle = GetWindowExStyle(hwndTarget);
    HWND hwndOwner = GetWindow(hwndTarget, GW_OWNER);

    g_hwndEnchanted = hwndTarget;

    style &= ~(WS_HSCROLL | WS_VSCROLL | WS_VISIBLE);

    TCHAR szText[MAX_PATH];
    lstrcpy(szText, g_szEnchantedText);
    lstrcat(szText, TEXT(" (Enchanted)"));

    g_hwndGhost = CreateWindowEx(exstyle, s_szGhostClass, szText, style,
        rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top,
        hwndOwner, NULL, GetModuleHandle(NULL), NULL);
    if (!g_hwndGhost)
    {
        MessageBox(NULL, TEXT("CreateWindowEx"), NULL, MB_ICONERROR);
        return FALSE;
    }

    g_hbmEnchanted = GetWindowBitmap(hwndTarget, rc.right - rc.left, rc.bottom - rc.top);
    SaveBitmapToFile(TEXT("Enchanted.bmp"), g_hbmEnchanted);
    Whiten32BppBitmap(g_hbmEnchanted);
    SaveBitmapToFile(TEXT("Whiten.bmp"), g_hbmEnchanted);

    HWND hwndPrev = GetWindow(g_hwndEnchanted, GW_HWNDPREV);
    ShowWindowAsync(g_hwndEnchanted, SW_HIDE);
    SetWindowPos(g_hwndGhost, hwndPrev, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_DRAWFRAME);
    MoveWindow(g_hwndGhost, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);
    SetWindowLong(g_hwndGhost, GWL_STYLE, style | WS_VISIBLE);
    InvalidateRect(g_hwndGhost, NULL, TRUE);

    SetTimer(g_hwndGhost, 999, 1000, NULL);

    MSG msg;
    while (GetMessage(&msg, NULL, 0, 0))
    {
        if (!IsWindow(g_hwndGhost) || !IsWindow(g_hwndEnchanted))
            break;
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    DoUnenchant();

    return TRUE;
}

BOOL OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
    TCHAR szText[128];
    wsprintf(szText, TEXT("This window HWND is 0x%p"), hwnd);
    SetDlgItemText(hwnd, stc1, szText);

    SetDlgItemText(hwnd, edt1, TEXT("0x"));
    return TRUE;
}

void OnOK(HWND hwnd)
{
    TCHAR szText[64];
    GetDlgItemText(hwnd, IDOK, szText, ARRAYSIZE(szText));
    if (lstrcmp(szText, TEXT("Unenchant")) == 0)
    {
        SetDlgItemText(hwnd, IDOK, TEXT("Enchant Now!"));
        DestroyWindow(g_hwndGhost);
        return;
    }

    GetDlgItemText(hwnd, edt1, szText, ARRAYSIZE(szText));

    HWND hwndTarget = (HWND)(ULONG_PTR)_tcstoul(szText, NULL, 0);

    if (IsWindow(hwndTarget))
    {
        SetDlgItemText(hwnd, IDOK, TEXT("Unenchant"));
        DoEnchant(hwndTarget);
        SetDlgItemText(hwnd, IDOK, TEXT("Enchant Now!"));
    }
    else
    {
        MessageBox(hwnd, TEXT("Invalid HWND"), NULL, MB_ICONERROR);
    }
}

void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
    switch (id)
    {
    case IDOK:
        OnOK(hwnd);
        break;
    case IDCANCEL:
        EndDialog(hwnd, id);
        break;
    }
}

INT_PTR CALLBACK
DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialog);
        HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
    }
    return 0;
}

INT WINAPI
WinMain(HINSTANCE   hInstance,
        HINSTANCE   hPrevInstance,
        LPSTR       lpCmdLine,
        INT         nCmdShow)
{
    DialogBox(hInstance, MAKEINTRESOURCE(1), NULL, DialogProc);
    return 0;
}
